#!/usr/local/bin/perl
# $Revision: 1.4 $$Date: 2001/04/18 20:15:38 $

# This tries to interpret a genbank file and heuristically
# turn it into something that pipmaker can use.  The -verbose
# flag is an undocumented feature that produces more verbose
# output for debugging purposes.

use 5.6.0;
use strict;
use warnings;
use Boulder::Genbank;

# Get Boulder from
#  http://stein.cshl.org/software/boulder/
#  http://bioperl.org/ScriptCentral/

my $argv0 = $0;
$argv0 =~ s{.*/}{};
my $filename = "";
my $verbose = 0;
my (@cds, @exon, @gene, @mrna, @prim);

sub strand
{
	return "<" if $_[0] =~ /complement/;
	return ">";
}

sub posns
{
	my $p = shift;
	my @x = ($p);

	if ( $p =~ /complement\((.+)\)/ ) {
		$p = $1;
		@x = ($p);
	}
	if ( $p =~ /join\((.+)\)/ ) {
		@x = split(/,/, $1);
	}
	map {s/[^\d]/ /g} @x;
	map {s/^ //g} @x;
	return @x;
}

sub leftpos
{
	my @p = posns($_[0]);
	$p[0] =~ /(\d+)\s+(\d+)/;
	return $1;
}

sub rightpos
{
	my @p = posns($_[0]);
	$p[$#p] =~ /(\d+)\s+(\d+)/;
	return $2;
}

sub summary
{
	print "#";
	print " ", scalar(@gene)," genes;";
	print " ", scalar(@exon)," exons;";
	print " ", scalar(@cds)," cds;";
	print " ", scalar(@mrna)," mrna;";
	print " ", scalar(@prim)," prim_transcript;";
	print "\n\n";
}

sub commentpos
{
	my ($A,$B,$C) = @_;
	print "# $A ", $B->Position(), "\n" if $B->Gene() && $B->Gene() eq $C;
}

sub strLRpos
{
	my ($A,$p,$C) = @_;
	return "$A ". leftpos($p) ." ". rightpos($p) ." $C\n";
}

sub overview
{
	my $n = shift;
	my $o = shift;
	my $g = $o->Gene;
	my $p = $o->Position;

	print "# $n $g ", $p, "\n";
	foreach (@exon) { commentpos "exon", $_, $g; }
	foreach (@mrna) { commentpos "mrna", $_, $g; }
	foreach (@cds)  { commentpos "cds ", $_, $g; }
	print "\n";
}

sub collect_no_gene
{
	my @v = ();
	foreach (@_) {
		push @v, $_ if (!defined $_->Gene || $_->Gene eq "");
	}
	return @v;
}

sub collect_for_gene
{
	my $g = shift;
	my @v = ();
	foreach (@_) {
		push @v, $_ if ($_->Gene && $_->Gene eq $g);
	}
	return @v;
}

sub report_gene
{
	print "# (report-gene mode)\n\n" if $verbose;
	foreach (@gene) {
		my $g = $_->Gene;
		my $p = $_->Position;

		overview "gene", $_;

		my $C = "";
		my $G = strLRpos strand($p), $p, $g;
		foreach (@cds) {
			if ($_->Gene && $_->Gene eq $g) {
				$C = strLRpos "+", $_->Position(), "";
				last;  # XXX - multiple cds?
			}
		}
		# There may not be cds or mrna for any given gene,
		# but use what is available.
		my @x = collect_for_gene($g, @mrna);
		if (@x) {
			print "# using $#x mrna\n" if $verbose;
		} else {
			@x = collect_for_gene($g, @cds);
			print "# using $#x cds\n" if $verbose;
		}

		foreach (@x) {
			print $G;
			print $C;
			print join("\n",posns($_->Position())), "\n";
			print "\n";
		}
	}
}

sub report_cds
{
	print "# (report-cds mode)\n\n" if $verbose;
	foreach (@cds) {
		my $p = $_->Position;
		my $g = $_->Product;

		print "# cds ", $p, "\n" if $verbose;
		print strLRpos(strand($p), $p, $g);
		print join("\n",posns($p)), "\n\n";
	}
}

sub report_strays
{
	print "# stray mRNA\n";
	foreach (collect_no_gene @mrna) {
		my $p = $_->Position;
		print strLRpos(strand($p), $p, "mRNA");
		print join("\n",posns($_->Position())), "\n\n";
	}
	print "# stray CDs\n";
	foreach (collect_no_gene @cds) {
		my $p = $_->Position;
		print strLRpos(strand($p), $p, $_->Product);
		print join("\n",posns($_->Position())), "\n\n";
	}
}

# --- main ---

my $narg = @ARGV;
if ($narg == 1) {
	$filename = $ARGV[0];
} elsif ($narg == 2 && $ARGV[1] eq '-verbose') {
	$filename = $ARGV[0];
	$verbose = 1;
} elsif ($narg == 2 && $ARGV[0] eq '-verbose') {
	$filename = $ARGV[1];
	$verbose = 1;
} else {
	print STDERR "Usage:  $argv0 genbank_file > output_file\n";
	exit 1;
}
if ($filename eq '-') { $filename = '/dev/stdin'; }

my $gb = new Boulder::Genbank(-accessor=>'File',-param=>$filename) or die;

while (my $s = $gb->get) {
	# Report the accession number.  Something else might be more useful.
	print $s->Accession, "\n\n";

	next unless $s && $s->Features;
	@cds = $s->Features->Cds;
	@exon = $s->Features->Exon;
	@gene = $s->Features->Gene;
	@mrna = $s->Features->Mrna;
	@prim = $s->Features->Prim_transcript;

	summary; # if $verbose;
	# The file may not list any genes.  If it does, enumerate those.
	# Otherwise, see if it lists any cds.
	if (@gene) {
		report_gene; report_strays;
	} else {
		report_cds;
	}
}

